/*
 * Copyright 2003-2018 The Music Player Daemon Project
 * http://www.musicpd.org
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include "FormatString.hxx"
#include "AllocatedString.hxx"

#include <stdio.h>
#include <stdlib.h>

AllocatedString<>
FormatStringV(const char *fmt, va_list args) noexcept
{
	va_list tmp;
	va_copy(tmp, args);
	const int length = vsnprintf(NULL, 0, fmt, tmp);
	va_end(tmp);

	if (length <= 0)
		/* wtf.. */
		abort();

	char *buffer = new char[length + 1];
	vsnprintf(buffer, length + 1, fmt, args);
	return AllocatedString<>::Donate(buffer);
}

AllocatedString<>
FormatString(const char *fmt, ...) noexcept
{
	va_list args;
	va_start(args, fmt);
	auto p = FormatStringV(fmt, args);
	va_end(args);
	return p;
}
