/*
 * Copyright (C) 2011-2012 Max Kellermann <max.kellermann@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * - Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * - Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the
 * distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * FOUNDATION OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef SLES_ENGINE_HPP
#define SLES_ENGINE_HPP

#include <SLES/OpenSLES.h>

namespace SLES {
	/**
	 * OO wrapper for an OpenSL/ES SLEngineItf variable.
	 */
	class Engine {
		SLEngineItf engine;

	public:
		Engine() = default;
		explicit Engine(SLEngineItf _engine):engine(_engine) {}

		SLresult CreateAudioPlayer(SLObjectItf *pPlayer,
					   SLDataSource *pAudioSrc, SLDataSink *pAudioSnk,
					   SLuint32 numInterfaces,
					   const SLInterfaceID *pInterfaceIds,
					   const SLboolean *pInterfaceRequired) {
			return (*engine)->CreateAudioPlayer(engine, pPlayer,
							    pAudioSrc, pAudioSnk,
							    numInterfaces, pInterfaceIds,
							    pInterfaceRequired);
		}

		SLresult CreateOutputMix(SLObjectItf *pMix,
					 SLuint32 numInterfaces,
					 const SLInterfaceID *pInterfaceIds,
					 const SLboolean *pInterfaceRequired) {
			return (*engine)->CreateOutputMix(engine, pMix,
							  numInterfaces, pInterfaceIds,
							  pInterfaceRequired);
		}
	};
}

#endif
